/*
 * SPDX-FileCopyrightText: 2017-2017 CSSlayer <wengxt@gmail.com>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */
#ifndef _FCITX_LIBIME_TABLE_CONSTANTS_H_
#define _FCITX_LIBIME_TABLE_CONSTANTS_H_

namespace libime {
constexpr int TABLE_AUTOPHRASE_SIZE = 256;
constexpr float TABLE_DEFAULT_MIN_DISTANCE = 1.0f;
} // namespace libime

#endif // _FCITX_LIBIME_TABLE_CONSTANTS_H_
