/*
 * SPDX-FileCopyrightText: 2017-2017 CSSlayer <wengxt@gmail.com>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */
#include "log.h"

namespace libime {
FCITX_DEFINE_LOG_CATEGORY(libime_table_logcategory, "libime-table");
}
