/*
 * SPDX-FileCopyrightText: 2017-2017 CSSlayer <wengxt@gmail.com>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */
#ifndef _FCITX_LIBIME_PINYIN_CONSTANTS_H_
#define _FCITX_LIBIME_PINYIN_CONSTANTS_H_

namespace libime {
constexpr float PINYIN_DISTANCE_PENALTY_FACTOR = 3;
}

#endif // _FCITX_LIBIME_PINYIN_CONSTANTS_H_
