/*
 * SPDX-FileCopyrightText: 2017-2017 CSSlayer <wengxt@gmail.com>
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#include "utils.h"

namespace libime {

FCITX_DEFINE_LOG_CATEGORY(libime_logcategory, "libime")
}
